package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.espec.model.ESpecStep;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.SfaExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.DesiredStatus;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.NonSfaExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.states.*;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.util.AggregateManagerWrapperFactory;
import be.iminds.ilabt.jfed.highlevel.util.LogEntryGeneratorWrappingLogger;
import be.iminds.ilabt.jfed.highlevel.util.ProxySocketFactoryProvider;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.AggregateManagerWrapper;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.RspecLink;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec_fx.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;
import java.util.stream.Collectors;

/**
 * This executes all actions that jFed needs to do AFTER an experiment is READY:
 *
 * - Sharing LANs (if specified in RSpec)
 * - Waiting for startup scripts (if wait method specified in RSpec)
 * - Distributing SSH keys (if specified in RSpec)
 * - Executing ansible (if ansible stuff specified in RSpec)
 * - Execute the dir, upload and execute steps of the ExperimentSpecification (if present)
 *
 * This does not:
 * - Test Connectivity  (should be done BEFORE this step)
 */
public class SetupSoftwareExperimentJob extends JobWithSshConnectionManager<Boolean> {
    private static final Logger ACTUAL_LOG = LoggerFactory.getLogger(SetupSoftwareExperimentJob.class);
    private final Logger LOG;

    @Nonnull private final AggregateManagerWrapperFactory aggregateManagerWrapperFactory;
    @Nonnull private final ExperimentPartControllerManager epcManager;

    //these 2 are used for replaying part of an ESpec.

    /** "Position" in the ESpec from which to start (inclusive), all previous steps are skipped. Do EVERY step (before stop) if this is null. */
    @Nullable private final ESpecStep startPosition;
    /** "Position" in the ESpec from which to stop (inclusive), all next steps are skipped. Do EVERY step (after start) if this is null. */
    @Nullable private final ESpecStep stopPosition;

    private final boolean runAll;

    private Boolean success = null;

    public SetupSoftwareExperimentJob(@Nonnull Experiment experiment,
                                      @Nonnull HighLevelTaskFactory hltf,
                                      @Nonnull TaskThread tt,
                                      @Nonnull JobStateFactory jobStateFactory,
                                      @Nonnull AggregateManagerWrapperFactory aggregateManagerWrapperFactory,
                                      @Nonnull ExperimentPartControllerManager epcManager,
                                      @Nonnull ProxySocketFactoryProvider proxySocketFactoryProvider,
                                      @Nonnull GeniUserProvider geniUserProvider,
                                      @Nullable ESpecStep startPosition,
                                      @Nullable ESpecStep stopPosition) {
        super("Setup Software for Experiment " + experiment.getName(), experiment, hltf, tt,
                jobStateFactory, proxySocketFactoryProvider, geniUserProvider);

        this.aggregateManagerWrapperFactory = aggregateManagerWrapperFactory;
        this.epcManager = epcManager;

        this.LOG = new LogEntryGeneratorWrappingLogger(ACTUAL_LOG, getJobReport());

        this.startPosition = startPosition;
        this.stopPosition = stopPosition;
        this.runAll = startPosition == null && stopPosition == null;
    }

    @Nullable
    public ESpecStep getStartPosition() {
        return startPosition;
    }

    @Nullable
    public ESpecStep getStopPosition() {
        return stopPosition;
    }

    public boolean isRunAll() {
        return runAll;
    }

    @Override
    public Boolean execute() throws Exception {
        if (!hasSfaParts()) {
            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecFail(experiment.getExperimentSpecification().getRspecs().get(0),
                        "assert failed: experiment.getExperimentSpecification() == null", null);
                experiment.getExperimentSpecificationLogger().firePostRspecAll(false, Collections.emptyList());
            }
            return true;
        }

        LOG.debug("SetupSoftwareExperimentJob started");

        //Assumption: everything is now successfully running and we have connectivity available

        if (experiment.getSliceOrNull() == null ||
                experiment.getSlice().getRequestRspec() == null ||
                experiment.getSlice().getRequestRspec().getModelRspec() == null) {
            LOG.warn("Problem with experiment.getSlice().getRequestRspec(). " +
                    "Will assume this is normal and ABORT software setup.");

            if (experiment.getExperimentSpecification() != null) {
                experiment.getExperimentSpecificationLogger().firePostRSpecSuccess(
                        experiment.getExperimentSpecification().getRspecs().get(0));
                experiment.getExperimentSpecificationLogger().firePostRspecAll(
                        true, getExperimentClientIds());
            }

            success = true;
            return true;
        }

        if (runAll) {
            boolean hasAutoShareLans = false;
            Collection<? extends RspecLink> links = experiment.getSlice()
                    .getRequestRspec().getModelRspec().getLinks();
            for (RspecLink link : links) {
                List<String> autoShareLanNames = link.getAutoShareLanNames();
                //noinspection ConstantConditions
                if (autoShareLanNames != null && !autoShareLanNames.isEmpty()) {
                    hasAutoShareLans = true;
                    break;
                }
            }
            if (hasAutoShareLans) {
                LOG.info("Experiment has auto shared LANs. Will share them.");
                final AutoShareLanPartsState autoShareLanPartsState = new AutoShareLanPartsState();
                setAndRunState(autoShareLanPartsState);

                if (autoShareLanPartsState.getStatus() != ExperimentTaskStatus.SUCCESS) {
                    LOG.info("Failure sharing LANs");

                    if (experiment.getExperimentSpecification() != null) {
                        experiment.getExperimentSpecificationLogger().firePostRSpecFail(
                                experiment.getExperimentSpecification().getRspecs().get(0),
                                "Failure sharing LANs", null);
                        experiment.getExperimentSpecificationLogger().firePostRspecAll(
                                false, getExperimentClientIds());
                    }

                    success = false;
                    return false;
                }

                LOG.debug("Experiment auto shared LANs have been shared. Will wait for ready again.");
                final WaitForOpStatusExperimentJob.UpdateSfaPartsUntilTargetOpStatus updateUntilReadyState2 =
                        new WaitForOpStatusExperimentJob.UpdateSfaPartsUntilTargetOpStatus(
                                jobStateFactory, this, LOG,
                                DesiredStatus.GENI_READY);
                setAndRunState(updateUntilReadyState2);

                if (updateUntilReadyState2.isPermanentFailure()) {
                    LOG.info("WaitForPartsReadyState (after auto shared LANs) reported permanent failure");

                    if (experiment.getExperimentSpecification() != null) {
                        experiment.getExperimentSpecificationLogger().firePostRSpecFail(
                                experiment.getExperimentSpecification().getRspecs().get(0),
                                "WaitForPartsReadyState (after auto shared LANs) reported permanent failure", null);
                        experiment.getExperimentSpecificationLogger().firePostRspecAll(
                                false, getExperimentClientIds());
                    }

                    success = false;
                    return false;
                }

                LOG.debug("All is ready after sharing LANs");
            } else
                LOG.debug("Experiment has NO auto shared LANs");


            if (!experiment.getKeypairs().isEmpty() && DistributeSshKeysState.hasDistributeSshKeyPair(experiment)) {
                LOG.debug("Experiment requires a distributed keypair. Will generate and distribute it.");
                setAndRunSshUsingState(jobStateFactory.createDistributeSshKeysState(this));
            } else {
                LOG.debug("Experiment does not require a distributed keypair.");
            }

            if (WaitForExecuteFinishedState.hasWaitForExecute(experiment)) {
                LOG.debug("Experiment requires waiting for some server executes to finish.");
                setAndRunSshUsingState(jobStateFactory.createWaitForExecuteFinishedState(this));
            } else {
                LOG.debug("Experiment does not require waiting for server execute.");
            }

            if (ExecuteAnsibleServicesState.hasAnsibleServices(experiment)) {
                LOG.debug("Experiment requires executing ansible scripts.");
                setAndRunSshUsingState(jobStateFactory.createExecuteAnsibleServicesState(this));
            } else {
                LOG.debug("Experiment does not require executing ansible scripts.");
            }
        }

        if (experiment.getExperimentSpecification() != null) {
            LOG.info("Experiment run with ExperimentSpecification. Will now start its dir, upload and execute phases.");

            experiment.getExperimentSpecificationLogger().firePostRSpecSuccess(
                    experiment.getExperimentSpecification().getRspecs().get(0));
            experiment.getExperimentSpecificationLogger().firePostRspecAll(
                    true, getExperimentClientIds());

            List<State> states = jobStateFactory.createOrderedExperimentServicesStates(this);
            for (State state: states) {
                setAndRunSshUsingState(state);
                if (state.getStatus() != ExperimentTaskStatus.SUCCESS) {
                    LOG.info("Experiment run without ExperimentSpecification. " +
                                    "Problem in ESpec phase {} (msg={}), got status={}. Will abort in failure.",
                            state.getName(), state.getMessage(), state.getStatus()
                            );
                    //INACTIVE, BUSY, FAILED, WARNING
                    experiment.getExperimentSpecificationLogger().firePostExecuteAll(false);
                    success = false;
                    return false;
                }
            }
            experiment.getExperimentSpecificationLogger().firePostExecuteAll(true);
        } else {
            LOG.info("Experiment run without ExperimentSpecification.");
        }

        LOG.debug("SetupSoftwareExperimentJob stopped successfully");

        success = true;
        return true;
    }

    @Nonnull
    private List<String> getExperimentClientIds() {
        if (experiment.getSliceOrNull() == null || experiment.getSliceOrNull().getRequestRspec() == null) {
            return Collections.emptyList();
        }
        RequestRspecSource requestRspecSource = experiment.getSliceOrNull().getRequestRspec();
        FXModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec(ModelRspecType.FX);
        assert model != null;

        if (model == null) {
            return Collections.emptyList();
        }

        return model.getNodes().stream().map(RspecNode::getClientId).filter(Objects::nonNull).collect(Collectors.toList());
    }

    public Boolean isSuccess() {
        return success;
    }

    public class AutoShareLanPartsState extends SlicedState<AutoShareLanPartStateSlice> {
        private final List<AutoShareLanPartStateSlice> slices;

        protected AutoShareLanPartsState() {
            super("Auto Sharing LANs in all ExperimentParts");

            slices = new ArrayList<>();

            List<SfaExperimentPart> sfaExperimentParts =
                    experiment.getPartsStream()
                            .filter(part -> part.getState() != InternalState.FAILED)
                            .filter(part -> part instanceof SfaExperimentPart)
                            .map(part -> (SfaExperimentPart) part)
                            .collect(Collectors.toList());

            experiment.requireSlice();
            if (experiment.getSlice().getRequestRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec().getLinks() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec().getLinks() is null which is not supposed to occur in this state");

            Collection<? extends RspecLink> links = experiment.getSlice().getRequestRspec().getModelRspec().getLinks();

            for (SfaExperimentPart sfaExperimentPart : sfaExperimentParts) {
                Server auth = sfaExperimentPart.getConnectSfaAuthority();
                boolean hasAutoShareLans = false;

                for (RspecLink link : links) {
                    List<String> autoShareLanNames = link.getAutoShareLanNames();
                    //noinspection ConstantConditions
                    if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getDefaultComponentManagerAsGeniUrn())) {
                        hasAutoShareLans = true;
                        break;
                    }
                }

                if (hasAutoShareLans) {
                    LOG.debug("Auto Sharing LAN(s) in " + auth.getName());
                    AutoShareLanPartStateSlice slice = new AutoShareLanPartStateSlice(sfaExperimentPart);
                    slices.add(slice);
                } else {
                    LOG.debug("No Auto Shared LAN  in " + auth.getName());
                }
            }

            LOG.debug("AutoShareLanPartsState constructor finished with " + slices.size() + " slices.");

            assert !slices.isEmpty();
        }

        @Override
        public Collection<AutoShareLanPartStateSlice> getSlices() {
            return slices;
        }
    }

    public class AutoShareLanPartStateSlice extends ExperimentPartStateSlice<SfaExperimentPart> {

        public AutoShareLanPartStateSlice(SfaExperimentPart experimentPart) {
            super(SetupSoftwareExperimentJob.this, experimentPart);
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            if (!supportsShareLan(experimentPart.getConnectSfaAuthority())) {
                LOG.warn("Share LAN on {} is not supported. Will try anyway.", experimentPart.getName());
                //return ExperimentTaskStatus.WARNING;
            }
            experiment.requireSlice();
            if (experiment.getSlice().getRequestRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec() is null which is not supposed to occur in this state");
            if (experiment.getSlice().getRequestRspec().getModelRspec().getLinks() == null)
                throw new IllegalStateException("experiment.getSlice().getRequestRspec().getModelRspec().getLinks() is null which is not supposed to occur in this state");

            Collection<? extends RspecLink> links = experiment.getSlice().getRequestRspec().getModelRspec().getLinks();

            ExperimentTaskStatus worstStatus = null;
            Server auth = experimentPart.getConnectSfaAuthority();
            LOG.debug("Start sharing LAN(s) on " + auth.getDefaultComponentManagerAsGeniUrn());
            for (RspecLink link : links) {
                List<String> autoShareLanNames = link.getAutoShareLanNames();
                //noinspection ConstantConditions
                if (autoShareLanNames != null && !autoShareLanNames.isEmpty() && link.getComponentManagerUrns().contains(auth.getDefaultComponentManagerAsGeniUrn())) {
                    String sharedLanName = autoShareLanNames.get(0);

                    if (autoShareLanNames.size() > 1)
                        LOG.warn("There is more than one auto share LAN name per link (for link\"" + link.getClientId() + "\"), ignoring all but first.");

                    LOG.debug("Sharing LAN " + link.getClientId() + " as " + sharedLanName + " on " + auth.getDefaultComponentManagerAsGeniUrn());

                    ShareLanAtAuthorityPartState shareLanState =
                            jobStateFactory.createShareLanState(SetupSoftwareExperimentJob.this, experimentPart, link.getClientId(), sharedLanName);

                    setAndRunState(shareLanState);
                    ExperimentTaskStatus status = shareLanState.getStatus();
                    LOG.debug("Shared LAN " + link.getClientId() + " as " + sharedLanName + " on " + auth.getDefaultComponentManagerUrn() + " -> status = " + status);

                    if (status != ExperimentTaskStatus.SUCCESS || worstStatus == null) {
                        worstStatus = status;
                    }
                }
            }

            //there should be at least one link shared, or this shouldn't have been called.
            assert worstStatus != null;

            LOG.debug("Stopped sharing LAN(s) on " + auth.getDefaultComponentManagerUrn() + " -> worst status = " + worstStatus);

            return worstStatus;
        }

        private boolean supportsShareLan(Server auth) {
            AggregateManagerWrapper amWrapper =
                    aggregateManagerWrapperFactory.getAggregateManagerWrapper(auth);

            return amWrapper.isShareLanSupported();
        }
    }



    private Collection<SfaExperimentPart> getSfaParts() {
        return experiment.getPartsStream()
                .filter(part -> part instanceof SfaExperimentPart)
                .map(SfaExperimentPart.class::cast)
                .collect(Collectors.toList());

    }

    private Collection<NonSfaExperimentPartController> getNonSfaExperimentPartControllers() {

        Set<NonSfaExperimentPartController> controllers = new HashSet<>();

        for (ExperimentPart experimentPart : experiment.getPartsListCopy()) {
            if (experimentPart instanceof SfaExperimentPart)
                continue;


            ExperimentPartController epc = epcManager.getController(experimentPart);
            if (!(epc instanceof NonSfaExperimentPartController))
                throw new RuntimeException("Expected a non-sfa ExperimentPartController");

            controllers.add((NonSfaExperimentPartController) epc);
        }

        return controllers;
    }

    private boolean hasNonSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> !(part instanceof SfaExperimentPart));
    }

    private boolean hasSfaParts() {
        return experiment.getPartsStream().anyMatch(part -> (part instanceof SfaExperimentPart));
    }
}
